@@ -1,7 +1,8 @@
 from geosoup.common import Sublist, Handler, Opt, FTPHandler, Logger, Timer
 from geosoup.regression import RFRegressor, MRegressor, HRFRegressor
 from geosoup.samples import Samples
-from geosoup.raster import Raster, MultiRaster, Terrain, GDAL_FIELD_DEF, GDAL_FIELD_DEF_INV
+from geosoup.raster import Raster, MultiRaster, GDAL_FIELD_DEF, GDAL_FIELD_DEF_INV
+from geosoup.terrain import Terrain
 from geosoup.distance import Mahalanobis, Distance, Euclidean
 from geosoup.exceptions import ObjectNotFound, UninitializedError, FieldError, \
     FileNotFound, TileNotFound, ImageProcessingError
