@@ -1,6 +1,7 @@
 import numpy as np
 from geosoup.common import Handler, Opt, Sublist
 import warnings
+import json
 from osgeo import gdal, gdal_array, ogr, osr, gdalconst
 np.set_printoptions(suppress=True)
 
@@ -11,7 +12,6 @@ gdal.AllRegister()
 
 __all__ = ['Raster',
            'MultiRaster',
-           'Terrain',
            'GDAL_FIELD_DEF',
            'GDAL_FIELD_DEF_INV']
 
@@ -98,6 +98,112 @@ class Raster(object):
         else:
             return "<raster with path {ras}>".format(ras=self.name,)
 
+    def initialize(self,
+                   get_array=False,
+                   band_order=None,
+                   finite_only=True,
+                   nan_replacement=0.0,
+                   verbose=False):
+
+        """
+        Initialize a raster object from a file
+        :param get_array: flag to include raster as 3 dimensional array (bool)
+        :param band_order: band location list of integers (starting at 0; ignored if get_array is False)
+        :param finite_only: flag to remove non-finite values from array (ignored if get_array is False)
+        :param nan_replacement: replacement for all non-finite replacements
+        :param verbose: If some steps should be displayed
+        (ignored if finite_only, get_array is false)
+        :return None
+        """
+        self.init = True
+
+        raster_name = self.name
+
+        if Handler(raster_name).file_exists() or 'vsimem' in self.name:
+            fileptr = gdal.Open(raster_name)  # open file
+            self.datasource = fileptr
+            self.metadict = Raster.get_raster_metadict(file_name=raster_name)
+
+        elif self.datasource is not None:
+            fileptr = self.datasource
+            self.metadict = Raster.get_raster_metadict(file_ptr=fileptr)
+
+        else:
+            raise ValueError('No datasource found')
+
+        if band_order is None:
+            band_order = list(range(fileptr.RasterCount))
+        elif max(band_order) >= fileptr.RasterCount:
+            raise ValueError('Band indices must be smaller than number of available bands')
+
+        self.shape = [fileptr.RasterCount if not get_array else len(band_order),
+                      fileptr.RasterYSize,
+                      fileptr.RasterXSize]
+
+        self.dtype = fileptr.GetRasterBand(1).DataType
+        self.nodatavalue = fileptr.GetRasterBand(1).GetNoDataValue()
+
+        # if get_array flag is true
+        if get_array:
+            if verbose:
+                Opt.cprint('Reading bands: {}'.format(" ".join([str(b) for b in band_order])))
+
+            nbands = len(band_order)
+
+            # initialize array
+            if self.array_offsets is None:
+                array3d = np.zeros((nbands,
+                                    self.shape[1],
+                                    self.shape[2]),
+                                   gdal_array.GDALTypeCodeToNumericTypeCode(self.dtype))
+            else:
+                array3d = np.zeros((nbands,
+                                    self.array_offsets[3],
+                                    self.array_offsets[2]),
+                                   gdal_array.GDALTypeCodeToNumericTypeCode(self.dtype))
+
+            # get band names
+            names = list()
+
+            # read array and store the band values and name in array
+            for i, b in enumerate(band_order):
+                bandname = fileptr.GetRasterBand(b + 1).GetDescription()
+
+                if verbose:
+                    Opt.cprint('Reading band {}'.format(bandname))
+                names.append(bandname)
+
+                if self.array_offsets is None:
+                    array3d[i, :, :] = fileptr.GetRasterBand(b + 1).ReadAsArray()
+                else:
+                    array3d[i, :, :] = fileptr.GetRasterBand(b + 1).ReadAsArray(*self.array_offsets)
+
+            # if flag for finite values is present
+            if finite_only:
+                if np.isnan(array3d).any() or np.isinf(array3d).any():
+                    array3d[np.isnan(array3d)] = nan_replacement
+                    array3d[np.isinf(array3d)] = nan_replacement
+                    if verbose:
+                        Opt.cprint("Non-finite values replaced with " + str(nan_replacement))
+                else:
+                    if verbose:
+                        Opt.cprint("Non-finite values absent in file")
+
+            # assign to empty class object
+            self.array = array3d
+            self.bnames = names
+
+            self.transform = fileptr.GetGeoTransform()
+            self.crs_string = fileptr.GetProjection()
+
+        else:
+            self.bnames = list(fileptr.GetRasterBand(i + 1).GetDescription()
+                               for i in range(self.shape[0]))
+            self.transform = fileptr.GetGeoTransform()
+            self.crs_string = fileptr.GetProjection()
+
+        self.bounds = self.get_bounds()
+
     def write_to_file(self,
                       outfile=None,
                       driver='GTiff',
@@ -227,7 +333,7 @@ class Raster(object):
                     try:
                         overviews_.append(int(elem))
                     except Exception as e:
-                        Opt.cprint('Conversion error: {} -for- {}'.format(e.args[0], elem))
+                        Opt.cprint('Conversion error: {} -for- {}'.format(e.args[1], elem))
 
                 overviews = overviews_
 
@@ -332,149 +438,6 @@ class Raster(object):
                                    driver=out_format,
                                    add_overviews=add_overviews)
 
-    def initialize(self,
-                   get_array=False,
-                   band_order=None,
-                   finite_only=True,
-                   nan_replacement=0.0,
-                   use_dict=None,
-                   sensor=None,
-                   verbose=False):
-
-        """
-        Initialize a raster object from a file
-        :param get_array: flag to include raster as 3 dimensional array (bool)
-        :param band_order: band location array (int starting at 0; ignored if get_array is False)
-        :param finite_only: flag to remove non-finite values from array (ignored if get_array is False)
-        :param nan_replacement: replacement for all non-finite replacements
-        :param use_dict: Dictionary to use for renaming bands
-        :param sensor: Sensor to be used with dictionary (resources.bname_dict)
-        :param verbose: If some steps should be displayed
-        (ignored if finite_only, get_array is false)
-        :return None
-        """
-        self.init = True
-
-        raster_name = self.name
-
-        if Handler(raster_name).file_exists() or 'vsimem' in self.name:
-            fileptr = gdal.Open(raster_name)  # open file
-            self.datasource = fileptr
-            self.metadict = Raster.get_raster_metadict(file_name=raster_name)
-
-        elif self.datasource is not None:
-            fileptr = self.datasource
-            self.metadict = Raster.get_raster_metadict(file_ptr=fileptr)
-
-        else:
-            raise ValueError('No datasource found')
-
-        # get shape metadata
-        bands = fileptr.RasterCount
-        rows = fileptr.RasterYSize
-        cols = fileptr.RasterXSize
-
-        # if get_array flag is true
-        if get_array:
-
-            # get band names
-            names = list()
-
-            # band order
-            if band_order is None:
-                array3d = fileptr.ReadAsArray()
-
-                # if flag for finite values is present
-                if finite_only:
-                    if np.isnan(array3d).any() or np.isinf(array3d).any():
-                        array3d[np.isnan(array3d)] = nan_replacement
-                        array3d[np.isinf(array3d)] = nan_replacement
-                        if verbose:
-                            Opt.cprint("Non-finite values replaced with " + str(nan_replacement))
-                    else:
-                        if verbose:
-                            Opt.cprint("Non-finite values absent in file")
-
-                # get band names
-                for i in range(0, bands):
-                    names.append(fileptr.GetRasterBand(i + 1).GetDescription())
-
-            # band order present
-            else:
-                if verbose:
-                    Opt.cprint('Reading bands: {}'.format(" ".join([str(b) for b in band_order])))
-
-                bands = len(band_order)
-
-                # bands in array
-                n_array_bands = len(band_order)
-
-                # initialize array
-                if self.array_offsets is None:
-                    array3d = np.zeros((n_array_bands,
-                                        rows,
-                                        cols),
-                                       gdal_array.GDALTypeCodeToNumericTypeCode(fileptr.GetRasterBand(1).DataType))
-                else:
-                    array3d = np.zeros((n_array_bands,
-                                        self.array_offsets[3],
-                                        self.array_offsets[2]),
-                                       gdal_array.GDALTypeCodeToNumericTypeCode(fileptr.GetRasterBand(1).DataType))
-
-                # read array and store the band values and name in array
-                for i, b in enumerate(band_order):
-                    bandname = fileptr.GetRasterBand(b + 1).GetDescription()
-
-                    if verbose:
-                        Opt.cprint('Reading band {}'.format(bandname))
-
-                    if self.array_offsets is None:
-                        array3d[i, :, :] = fileptr.GetRasterBand(b + 1).ReadAsArray()
-                    else:
-                        array3d[i, :, :] = fileptr.GetRasterBand(b + 1).ReadAsArray(*self.array_offsets)
-
-                    names.append(bandname)
-
-                # if flag for finite values is present
-                if finite_only:
-                    if np.isnan(array3d).any() or np.isinf(array3d).any():
-                        array3d[np.isnan(array3d)] = nan_replacement
-                        array3d[np.isinf(array3d)] = nan_replacement
-                        if verbose:
-                            Opt.cprint("Non-finite values replaced with " + str(nan_replacement))
-                    else:
-                        if verbose:
-                            Opt.cprint("Non-finite values absent in file")
-
-            # assign to empty class object
-            self.array = array3d
-            self.bnames = names
-            self.shape = [bands, rows, cols]
-            self.transform = fileptr.GetGeoTransform()
-            self.crs_string = fileptr.GetProjection()
-            self.dtype = fileptr.GetRasterBand(1).DataType
-
-        # if get_array is false
-        else:
-            # get band names
-            names = list()
-            for i in range(0, bands):
-                names.append(fileptr.GetRasterBand(i + 1).GetDescription())
-
-            # assign to empty class object without the array
-            self.bnames = names
-            self.shape = [bands, rows, cols]
-            self.transform = fileptr.GetGeoTransform()
-            self.crs_string = fileptr.GetProjection()
-            self.dtype = fileptr.GetRasterBand(1).DataType
-            self.nodatavalue = fileptr.GetRasterBand(1).GetNoDataValue()
-
-        self.bounds = self.get_bounds()
-
-        # remap band names
-        if use_dict is not None:
-            self.bnames = [use_dict[sensor][b] for b in self.bnames]
-
     def set_nodataval(self,
                       in_nodataval=255,
                       out_nodataval=0,
@@ -663,28 +626,90 @@ class Raster(object):
 
         return meta_dict
 
-    def change_type(self,
-                    out_type='int16'):
+    def get_stats(self,
+                  verbose=False,
+                  approx=False):
+
+        """
+        Method to compute statistics of the raster object, and store as raster property
+        :param verbose: If the statistics should be printed to console
+        :param approx: If approx statistics should be calculated instead to gain speed
+        :return: None
+        """
+
+        for ib in range(self.shape[0]):
+            band = self.datasource.GetRasterBand(ib+1)
+            band.ComputeStatistics(approx)
+            band_stats = dict(zip(['min', 'max', 'mean', 'stddev'], band.GetStatistics(int(approx), 0)))
+
+            if verbose:
+                Opt.cprint('Band {} : {}'.format(self.bnames[ib],
+                                                 str(band_stats)))
+
+            self.stats[self.bnames[ib]] = band_stats
+
+    def rescale(self,
+                out_type=None,
+                rescale=False,
+                scale_vals=None,
+                scale_approx=False):
 
         """
         Method to change the raster data type
-        :param out_type: Out data type. Options: int, int8, int16, int32, int64,
-                                                float, float, float32, float64,
-                                                uint, uint8, uint16, uint32, etc.
+        :param out_type: Out data type.
+                         Valid options: byte, int, long, float, double
+                                        int8, int16, int32, int64,
+                                        float32, float64,
+                                        uint8, uint16, uint32
+        :param rescale: If the image should be rescaled
+        :param scale_vals: Tuple of min and max values for rescaling
+        :param scale_approx: If approx stats should be calculated
         :return: None
         """
+
+        if out_type is None:
+            out_type = gdal_array.GDALTypeCodeToNumericTypeCode(self.dtype)
+        else:
+            common_types = {'byte': 'uint8',
+                            'int': 'int16',
+                            'float': 'float32',
+                            'double': 'float64',
+                            'long': 'int32'}
+
+            if out_type in common_types:
+                out_type = common_types[out_type]
+
         if gdal_array.NumericTypeCodeToGDALTypeCode(np.dtype(out_type)) != self.dtype:
 
-            self.array = self.array.astype(out_type)
+            if rescale:
+                if scale_vals is not None:
+
+                    if self.stats is None:
+                        self.get_stats(approx=scale_approx)
+                    elif len(self.stats) < self.shape[0]:
+                        self.get_stats(approx=scale_approx)
+
+                    out_arr = np.zeros(self.shape, dtype=out_type)
+
+                    for band_indx in range(self.shape[0]):
+                        min_val, max_val = self.stats[band_indx]['min'], self.stats[band_indx]['max']
+
+                        mdiff = float(scale_vals[1] - scale_vals[0]) / float(max_val - min_val)
+
+                        out_arr[band_indx, :, :] = ((self.array[band_indx, :, :].astype(np.float64) - min_val) * mdiff +
+                                                    scale_vals[0]).astype(out_type)
+
+                    self.array = out_arr
+                else:
+                    self.array = self.array.astype(out_type)
+            else:
+                self.array = self.array.astype(out_type)
+
             self.dtype = gdal_array.NumericTypeCodeToGDALTypeCode(self.array.dtype)
 
             if self.nodatavalue is not None:
                 self.nodatavalue = np.array(self.nodatavalue).astype(out_type).item()
 
-            print('Changed raster data type to {}\n'.format(out_type))
-        else:
-            print('Raster data type already {}\n'.format(out_type))
-
     def make_polygon_geojson_feature(self):
         """
         Make a feature geojson for the raster using its metaDict data
@@ -1375,28 +1400,6 @@ class Raster(object):
 
         return out_geom_extract
 
-    def get_stats(self,
-                  print_stats=False,
-                  approx=False):
-
-        """
-        Method to compute statistics of the raster object, and store as raster property
-        :param print_stats: If the statistics should be printed to console
-        :param approx: If approx statistics should be calculated instead to gain speed
-        :return: None
-        """
-
-        for ib in range(self.shape[0]):
-            band = self.datasource.GetRasterBand(ib+1)
-            band.ComputeStatistics(approx)
-            band_stats = dict(zip(['min', 'max', 'mean', 'stddev'], band.GetStatistics(int(approx), 0)))
-
-            if print_stats:
-                Opt.cprint('Band {} : {}'.format(self.bnames[ib],
-                                                 str(band_stats)))
-
-            self.stats[self.bnames[ib]] = band_stats
-
     def clip(self,
              cutline_file=None,
              cutline_blend=0,
@@ -1616,10 +1619,12 @@ class MultiRaster:
     def get_extent(self,
                    index=None,
                    intersection=True,
+                   verbose=False,
                    _return=True):
         """
         Method to get intersecting bounds of the raster objects
         :param index: list of indices of raster files/objects
+        :param verbose: If some of the steps should be displayed
         :param intersection: If intersection (True) should be returned or union (False) (Default: True)
         :param _return: Should the method return the bound coordinates
         :return: coordinates of intersection (minx, miny, maxx, maxy)
@@ -1646,21 +1651,24 @@ class MultiRaster:
             for geom in geoms[1:]:
                 temp_geom = temp_geom.Intersection(geom)
         else:
-            multi_geom = ogr.Geometry(ogr.wkbMultiPolygon)
+            temp_geom = ogr.Geometry(ogr.wkbMultiPolygon)
 
             for geom in geoms:
-                multi_geom.AddGeometryDirectly(geom)
+                temp_geom.AddGeometryDirectly(geom)
 
-            temp_geom = multi_geom.UnionCascaded()
+        temp_geom_json = temp_geom.ExportToJson()
+        json_str = json.loads(temp_geom_json)
 
-        temp_geom_wkt = temp_geom.ExportToWkt()
-
-        temp_coords = list(list(float(elem.strip()) for elem in elem_.strip()
-                                                                     .split(' '))
-                           for elem_ in temp_geom_wkt.replace('POLYGON', '')
-                                                     .replace('((', '')
-                                                     .replace('))', '')
-                                                     .split(','))
+        temp_coords = []
+        for level1_list in json_str['coordinates']:
+            for level2_list in level1_list:
+                if json_str['type'] == 'Polygon':
+                    temp_coords.append(level2_list)
+                elif json_str['type'] == 'MultiPolygon':
+                    for level3_list in level2_list:
+                        temp_coords.append(level3_list)
+                else:
+                    temp_coords.append(None)
 
         minx = min(list(coord[0] for coord in temp_coords))
         miny = min(list(coord[1] for coord in temp_coords))
@@ -1670,6 +1678,11 @@ class MultiRaster:
 
         self.intersection = (minx, miny, maxx, maxy)
 
+        if verbose:
+            Opt.cprint('xmin: {}, ymin: {}, xmax: {}, ymax: {}'.format(*[str(elem)
+                                                                         for elem in
+                                                                         self.intersection]))
+
         if _return:
             return self.intersection
 
@@ -1736,6 +1749,15 @@ class MultiRaster:
         vrt_dict['separate'] = True
         vrt_dict['hideNodata'] = False
 
+        vrt_dict['creationOptions'] = []
+
+        if 'compress' in kwargs:
+            vrt_dict['creationOptions'].append('COMPRESS={}'.format(str(kwargs['compress']).upper()))
+
+        if 'bigtiff' in kwargs:
+            vrt_dict['creationOptions'].append('BIGTIFF={}'.format(str(kwargs['bigtiff']).upper()))
+
+
         if verbose:
             Opt.cprint('Files: \n{}'.format('\n'.join(list(self.filelist[i] for i in order))))
 
@@ -1882,6 +1904,45 @@ class MultiRaster:
                        valid warp options in kwargs :
                        https://gdal.org/python/osgeo.gdal-module.html#WarpOptions
         :return: None
+        todo: 1) band selection: Use VRT to get bands that will be used at runtime
+              2) blending options: image bounds buffered at intersection with other image
+              using:
+                from scipy.ndimage.morphology import binary_erosion
+                from scipy.spatial.distance import cdist
+                def dist_from_edge(img):
+                    I = binary_erosion(img) # Interior mask
+                    C = img - I             # Contour mask
+                    out = C.astype(int)     # Setup o/p and assign cityblock distances
+                    out[I] = cdist(np.argwhere(C), np.argwhere(I), 'cityblock').min(0) + 1
+                    return out
+                    OR
+                    import numpy as np
+                    from scipy.ndimage import distance_transform_cdt
+                    def distance_from_edge(x):
+                        x = np.pad(x, 1, mode='constant')
+                        dist = distance_transform_cdt(x, metric='taxicab')
+                        return dist[1:-1, 1:-1]
+                    OR
+                    import numpy as np
+                    from scipy.spatial.distance import cdist
+                    def feature_dist(input):
+                        # Takes a labeled array as returned by scipy.ndimage.label and
+                        # returns an intra-feature distance matrix.
+                        I, J = np.nonzero(input)
+                        labels = input[I,J]
+                        coords = np.column_stack((I,J))
+                        sorter = np.argsort(labels)
+                        labels = labels[sorter]
+                        coords = coords[sorter]
+                        sq_dists = cdist(coords, coords, 'sqeuclidean')
+                        start_idx = np.flatnonzero(np.r_[1, np.diff(labels)])
+                        nonzero_vs_feat = np.minimum.reduceat(sq_dists, start_idx, axis=1)
+                        feat_vs_feat = np.minimum.reduceat(nonzero_vs_feat, start_idx, axis=0)
+                        return np.sqrt(feat_vs_feat)
+                        This approach requires O(N2) memory,
+                        where N is the number of nonzero pixels.
+                        If this is too demanding,
+                        you could "de-vectorize" it along one axis (add a for-loop).
         """
 
         if order is None:
@@ -1929,6 +1990,7 @@ class MultiRaster:
             Opt.cprint('Getting bounds ...')
 
         warp_dict['outputBounds'] = self.get_extent(index=order,
+                                                    verbose=verbose,
                                                     intersection=False)
         warp_dict['targetAlignedPixels'] = True
         warp_dict['outputType'] = self.rasters[0].dtype
@@ -1942,6 +2004,8 @@ class MultiRaster:
             warp_dict['cutlineLayer'] = blend_layer_name
             warp_dict['cutlineBlend'] = blend_pixels
 
+        warp_dict['creationOptions'] = []
+
         if 'compress' in kwargs:
             warp_dict['creationOptions'].append('COMPRESS={}'.format(str(kwargs['compress']).upper()))
 
@@ -1964,339 +2028,3 @@ class MultiRaster:
         if return_datasource:
             return warp_datasource
 
-
-class Terrain(Raster):
-    """Class to process DEM rasters"""
-
-    def __init__(self,
-                 name,
-                 array=None,
-                 bnames=None,
-                 metadict=None,
-                 dtype=None,
-                 shape=None,
-                 transform=None,
-                 crs_string=None):
-
-        super(Terrain, self).__init__(name,
-                                      array,
-                                      bnames,
-                                      metadict,
-                                      dtype,
-                                      shape,
-                                      transform,
-                                      crs_string)
-
-    def __repr__(self):
-        return 'Terrain: ' + super(Terrain, self).__repr__()
-
-    def slope(self,
-              outfile=None,
-              slope_format='degree',
-              file_format='GTiff',
-              compute_edges=True,
-              band=1,
-              scale=None,
-              algorithm='ZevenbergenThorne',
-              **creation_options):
-
-        """
-        Method to calculate slope
-        :param outfile: output file name
-        :param slope_format: format of the slope raster (valid options: 'degree', or 'percent')
-        :param file_format: Output file format (default: 'GTiff')
-        :param compute_edges: If the edges of the raster should be computed as well.
-                              This can present incomplete results at the edges and resulting
-                              rasters may show edge effects on mosaicking
-        :param band: Band index to use (default: 0)
-        :param scale: ratio of vertical to horizontal units
-        :param algorithm: slope algorithm to use
-                          valid options:
-                             4-neighbor: 'ZevenbergenThorne'
-                             8-neighbor: 'Horn'
-        :param creation_options: Valid creation options examples:
-                                 compress='LZW'
-                                 bigtiff='yes'
-        """
-        if not self.init:
-            self.initialize()
-
-        creation_option_list = list()
-        for key, value in creation_options.items():
-            creation_option_list.append('{}={}'.format(str(key).upper(), str(value).upper()))
-
-        slope_opts = gdal.DEMProcessingOptions(format=file_format,
-                                               computeEdges=compute_edges,
-                                               alg=algorithm,
-                                               slopeFormat=slope_format,
-                                               band=band,
-                                               scale=scale,
-                                               creationOptions=creation_option_list)
-
-        res = gdal.DEMProcessing(outfile,
-                                 self.datasource,
-                                 'slope',
-                                 options=slope_opts)
-        res = None
-
-    def aspect(self,
-               outfile=None,
-               file_format='GTiff',
-               compute_edges=True,
-               band=1,
-               scale=None,
-               algorithm='ZevenbergenThorne',
-               zero_for_flat=True,
-               trigonometric=False,
-               **creation_options):
-
-        """
-        Method to calculate aspect
-        :param outfile: output file name
-        :param file_format: Output file format (default: 'GTiff')
-        :param compute_edges: If the edges of the raster should be computed as well.
-                              This can present incomplete results at the edges and resulting
-                              rasters may show edge effects on mosaicking
-        :param band: Band index to use (default: 0)
-        :param scale: ratio of vertical to horizontal units
-        :param algorithm: slope algorithm to use
-                          valid options:
-                             4-neighbor: 'ZevenbergenThorne'
-                             8-neighbor: 'Horn'
-                             
-        :param zero_for_flat: whether to return 0 for flat areas with slope=0, instead of -9999.
-        :param trigonometric: whether to return trigonometric angle instead of azimuth.
-                             Here 0deg will mean East, 90deg North, 180deg West, 270deg South.
-        :param creation_options: Valid creation options examples:
-                                 compress='LZW'
-                                 bigtiff='yes'
-        """
-        if not self.init:
-            self.initialize()
-
-        creation_option_list = list()
-        for key, value in creation_options.items():
-            creation_option_list.append('{}={}'.format(str(key), str(value)))
-
-        aspect_opts = gdal.DEMProcessingOptions(format=file_format,
-                                                computeEdges=compute_edges,
-                                                creationOptions=creation_option_list,
-                                                alg=algorithm,
-                                                band=band,
-                                                scale=scale,
-                                                zeroForFlat=zero_for_flat,
-                                                trigonometric=trigonometric)
-
-        res = gdal.DEMProcessing(outfile,
-                                 self.datasource,
-                                 'aspect',
-                                 options=aspect_opts)
-        res = None
-
-    def tpi(self,
-            outfile=None,
-            file_format='GTiff',
-            compute_edges=True,
-            band=1,
-            scale=None,
-            algorithm='Horn',
-            **creation_options):
-
-        """
-        Method to calculate topographic position index
-        :param outfile: output file name
-        :param file_format: Output file format (default: 'GTiff')
-        :param compute_edges: If the edges of the raster should be computed as well.
-                              This can present incomplete results at the edges and resulting
-                              rasters may show edge effects on mosaicking
-        :param band: Band index to use (default: 0)
-        :param scale: ratio of vertical to horizontal units
-        :param algorithm: slope algorithm to use
-                          valid options:
-                             4-neighbor: 'ZevenbergenThorne'
-                             8-neighbor: 'Horn'
-        :param creation_options: Valid creation options examples:
-                                 compress='LZW'
-                                 bigtiff='yes'
-        """
-        if not self.init:
-            self.initialize()
-
-        creation_option_list = list()
-        for key, value in creation_options.items():
-            creation_option_list.append('{}={}'.format(str(key), str(value)))
-
-        tpi_opts = gdal.DEMProcessingOptions(format=file_format,
-                                             computeEdges=compute_edges,
-                                             creationOptions=creation_option_list,
-                                             band=band,
-                                             scale=scale,
-                                             alg=algorithm)
-
-        res = gdal.DEMProcessing(outfile,
-                                 self.datasource,
-                                 'TPI',
-                                 options=tpi_opts)
-        res = None
-
-    def tri(self,
-            outfile=None,
-            file_format='GTiff',
-            compute_edges=True,
-            band=1,
-            scale=None,
-            algorithm='Horn',
-            **creation_options):
-
-        """
-        Method to calculate topographic roughness index
-        :param outfile: output file name
-        :param file_format: Output file format (default: 'GTiff')
-        :param compute_edges: If the edges of the raster should be computed as well.
-                              This can present incomplete results at the edges and resulting
-                              rasters may show edge effects on mosaicking
-        :param band: Band index to use (default: 0)
-        :param scale: ratio of vertical to horizontal units
-        :param algorithm: slope algorithm to use
-                          valid options:
-                             4-neighbor: 'ZevenbergenThorne'
-                             8-neighbor: 'Horn'
-        :param creation_options: Valid creation options examples:
-                                 compress='LZW'
-                                 bigtiff='yes'
-        """
-        if not self.init:
-            self.initialize()
-
-        creation_option_list = list()
-        for key, value in creation_options.items():
-            creation_option_list.append('{}={}'.format(str(key), str(value)))
-
-        tpi_opts = gdal.DEMProcessingOptions(format=file_format,
-                                             computeEdges=compute_edges,
-                                             creationOptions=creation_option_list,
-                                             band=band,
-                                             scale=scale,
-                                             alg=algorithm)
-
-        res = gdal.DEMProcessing(outfile,
-                                 self.datasource,
-                                 'TRI',
-                                 options=tpi_opts)
-        res = None
-
-    def roughness(self,
-                  outfile=None,
-                  file_format='GTiff',
-                  compute_edges=True,
-                  band=1,
-                  scale=None,
-                  algorithm='ZevenbergenThorne',
-                  **creation_options):
-
-        """
-        Method to calculate DEM roughness
-        :param outfile: output file name
-        :param file_format: Output file format (default: 'GTiff')
-        :param compute_edges: If the edges of the raster should be computed as well.
-                              This can present incomplete results at the edges and resulting
-                              rasters may show edge effects on mosaicking
-        :param band: Band index to use (default: 0)
-        :param scale: ratio of vertical to horizontal units
-        :param algorithm: slope algorithm to use
-                          valid options:
-                             4-neighbor: 'ZevenbergenThorne'
-                             8-neighbor: 'Horn'
-        :param creation_options: Valid creation options examples:
-                                 compress='LZW'
-                                 bigtiff='yes'
-        """
-        if not self.init:
-            self.initialize()
-
-        creation_option_list = list()
-        for key, value in creation_options.items():
-            creation_option_list.append('{}={}'.format(str(key), str(value)))
-
-        tpi_opts = gdal.DEMProcessingOptions(format=file_format,
-                                             computeEdges=compute_edges,
-                                             creationOptions=creation_option_list,
-                                             band=band,
-                                             scale=scale,
-                                             alg=algorithm)
-
-        res = gdal.DEMProcessing(outfile,
-                                 self.datasource,
-                                 'Roughness',
-                                 options=tpi_opts)
-        res = None
-
-    def hillshade(self,
-                  outfile=None,
-                  file_format='GTiff',
-                  compute_edges=True,
-                  band=1,
-                  scale=None,
-                  algorithm='ZevenbergenThorne',
-                  z_factor=1,
-                  azimuth=315,
-                  altitude=90,
-                  combined=False,
-                  multi_directional=False,
-                  **creation_options):
-
-        """
-        Method to calculate DEM hillshade raster
-        :param outfile: output file name
-        :param file_format: Output file format (default: 'GTiff')
-        :param compute_edges: If the edges of the raster should be computed as well.
-                              This can present incomplete results at the edges and resulting
-                              rasters may show edge effects on mosaicking
-        :param band: Band index to use (default: 0)
-        :param scale: ratio of vertical to horizontal units
-        :param algorithm: slope algorithm to use
-                          valid options:
-                             4-neighbor: 'ZevenbergenThorne'
-                             8-neighbor: 'Horn'
-
-        :param z_factor: vertical exaggeration used to pre-multiply the elevations. (default: 1)
-        :param azimuth:  azimuth of the light, in degrees. (default: 315)
-                         0 if it comes from the top of the raster, 90 from the east and so on.
-                         The default value, 315, should rarely be changed
-                         as it is the value generally used to generate shaded maps.
-
-        :param altitude: altitude of the light, in degrees. (default: 90)
-                         90 if the light comes from above the DEM, 0 if it is raking light.
-        :param combined:  whether to compute combined shading,
-                         a combination of slope and oblique shading. (Default: False)
-        :param multi_directional: whether to compute multi-directional shading (Default: False)
-
-        :param creation_options: Valid creation options examples:
-                                 compress='LZW'
-                                 bigtiff='yes'
-        """
-        if not self.init:
-            self.initialize()
-
-        creation_option_list = list()
-        for key, value in creation_options.items():
-            creation_option_list.append('{}={}'.format(str(key), str(value)))
-
-        tpi_opts = gdal.DEMProcessingOptions(format=file_format,
-                                             computeEdges=compute_edges,
-                                             creationOptions=creation_option_list,
-                                             band=band,
-                                             scale=scale,
-                                             alg=algorithm,
-                                             zFactor=z_factor,
-                                             azimuth=azimuth,
-                                             altitude=altitude,
-                                             combined=combined,
-                                             multiDirectional=multi_directional)
-
-        res = gdal.DEMProcessing(outfile,
-                                 self.datasource,
-                                 'hillshade',
-                                 options=tpi_opts)
-        res = None
-
